/*
 *	mkgbafs.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#ifndef BOOL
typedef int	BOOL;
#endif
#include <gbafs.h>

#define CMDNAME		"mkgbafs"

static char	*Usage[] = {
	"Usage: %s [options] {file system root}\n",
	"options:	-s {system GBA file}\n",
	"		-o {output GBA file}\n",
	NULL,
};

#define ROUNDUPFS(x)	(((x) + ROMFS_CHUNK - 1) / ROMFS_CHUNK * ROMFS_CHUNK)

#define OFF2ROMADDR(o)	((o) + ROM_ADDR)

typedef struct _DIROBJ {
	PROMDIRENT	pdir;
	char *		pname;
	int		size;
} DIROBJ, *PDIROBJ;

static FILE	*sysfp;
static FILE	*outfp;

static char	*sysGBAfile = "system.gba";
static char	*outGBAfile = "output.gba";
static char	fsrootPath[_MAX_PATH];

static long	fsheadOffset;
static long	fsCurOffset;

static ROMFSHEAD	fshead;

#define MAX_BUFFER	4096

static char	buffer[MAX_BUFFER];

/*
 *	usage
 */
static void usage()
{
	char		**p;

	p = Usage;
	fprintf(stderr, *p++, CMDNAME);
	while (*p)
		fprintf(stderr, "%s", *p++);
	exit(1);
}

/*
 *	error
 */
static void error(char *fmt, void *p)
{
	fprintf(stderr, "%s: ", CMDNAME);
	fprintf(stderr, fmt, p);
	fprintf(stderr, "\n");
	exit(1);
}

/*
 *	checksize
 */
static void checksize(long size)
{
	if (ftell(outfp) + size > MAX_ROMFS_SIZE)
		error("No space.", NULL);
}

/*
 *	WriteFSData
 */
static void WriteFSData(void *bp, long size)
{
	checksize(size);
	if (fwrite(bp, size, 1, outfp) != 1)
		perror(outGBAfile), exit(1);
	fsCurOffset += size;
}

/*
 *	fillfile
 */
static void fillfile(long size)
{
	long		rsize;

	rsize = ROUNDUPFS(size) - size;
	if (rsize > 0) {
		checksize(rsize);
		if (fseek(outfp, rsize, SEEK_CUR))
			perror(outGBAfile), exit(1);
		fsCurOffset += rsize;
	}
}

/*
 *	copyfile
 */
static void copyfile(char *name, FILE *fp, long size)
{
	long		n, count;

	checksize(size);

	count = size;
	while (count > 0) {
		n = min(count, MAX_BUFFER);
		if (fread(buffer, n, 1, fp) != 1)
			perror(name), exit(1);
		WriteFSData(buffer, n);
		count -= n;
	}

	fillfile(size);
}

/*
 *	InitDirObj
 */
static void InitDirObj(PDIROBJ pdo, char *pname)
{
	pdo->pname = pname;
	pdo->pdir = NULL;
	pdo->size = 0;
}

/*
 *	AllocDirObj
 */
static PROMDIRENT AllocDirObj(PDIROBJ pdo)
{
	int		offset;
	PROMDIRENT	pdir;

	if (!pdo->pdir) {
		pdo->size = sizeof(*pdo->pdir);
		if (!(pdo->pdir = malloc(pdo->size)))
			perror(CMDNAME), exit(1);
		memset(pdo->pdir, 0, pdo->size);
		return pdo->pdir;
	} else {
		offset = pdo->size;
		pdo->size += sizeof(*pdo->pdir);
		pdo->pdir = realloc(pdo->pdir, pdo->size);
		if (!pdo->pdir)
			perror(CMDNAME), exit(1);
		pdir = (PROMDIRENT)((char *)pdo->pdir + offset);
		memset(pdir, 0, sizeof(*pdir));
		return pdir;
	}

}

/*
 *	FreeDirObj
 */
static void FreeDirObj(PDIROBJ pdo)
{
	if (pdo->pdir) free(pdo->pdir);
	InitDirObj(pdo, NULL);
}

/*
 *	BuildDirObj
 */
static void BuildDirObj(PROMDIRENT pdir, PDIROBJ pdo)
{
	long		fhoff;
	int		nsize;
	char *		pname;
	time_t		tm;

	/*
	 *	Build directory
	 */

	fhoff = fsCurOffset;
	nsize = pdo->pname ? ROUNDUPFS(strlen(pdo->pname) + 1) : 0;

	if (pdo->pname) {
		nsize = ROUNDUPFS(strlen(pdo->pname) + 1);
		if (!(pname = malloc(nsize)))
			perror(CMDNAME), exit(1);
		memset(pname, 0, nsize);

		strcpy(pname, pdo->pname);
		WriteFSData(pname, nsize);

		free(pname);
	} else
		nsize = 0;

	if (pdo->size)
		WriteFSData(pdo->pdir, pdo->size);

	/*
	 *	Build directory entry
	 */

	time(&tm);

	pdir->attr = DA_DIR;
	pdir->namesize = nsize;
	pdir->fileaddr = (u8 *)OFF2ROMADDR(fhoff);
	pdir->filesize = pdo->size;
	pdir->filetime = tm;
}

/*
 *	BuildFile
 */
static void BuildFile(PROMDIRENT pdir, struct _finddata_t *pfi)
{
	long		fhoff;
	int		nsize;
	FILE		*fp;
	char		*pname;

	/*
	 *	Build file header
	 */

	fhoff = fsCurOffset;

	nsize = ROUNDUPFS(strlen(pfi->name) + 1);
	if (!(pname = malloc(nsize)))
		perror(CMDNAME), exit(1);
	memset(pname, 0, nsize);

	strcpy(pname, pfi->name);
	WriteFSData(pname, nsize);

	/*
	 *	Build file
	 */

	if (!(fp = fopen(fsrootPath, "rb")))
		perror(fsrootPath), exit(1);
	copyfile(fsrootPath, fp, pfi->size);
	fclose(fp);

	/*
	 *	Build directory entry
	 */

	pdir->attr = DA_FILE;
	pdir->namesize = nsize;
	pdir->fileaddr = (u8 *)OFF2ROMADDR(fhoff);
	pdir->filesize = pfi->size;
	pdir->filetime = pfi->time_write;

	free(pname);
}

/*
 *	WalkDir
 */
static void WalkDir(PDIROBJ pdo, char *pname, int nestlevel)
{
	DIROBJ			sdo;
	PROMDIRENT		pdir;
	struct _finddata_t	fi;
	int			fh;
	char			*plast;

	InitDirObj(pdo, pname);
	plast = fsrootPath + strlen(fsrootPath);
	if (plast[-1] != '\\')
		*plast++ = '\\';
	strcpy(plast, "*.*");

	if ((fh = _findfirst(fsrootPath, &fi)) < 0)
		goto out;

	do {
		if (strcmp(fi.name, ".") == 0 || strcmp(fi.name, "..") == 0)
			continue;
		printf("%*c%s%c\n", nestlevel * 2, ' ', fi.name,
			(fi.attrib & _A_SUBDIR) ? '\\' : ' ');
		sprintf(plast, "%s", fi.name);
		pdir = AllocDirObj(pdo);
		if (fi.attrib & _A_SUBDIR) {
			WalkDir(&sdo, fi.name, nestlevel + 1);
			BuildDirObj(pdir, &sdo);
			FreeDirObj(&sdo);
		} else {
			BuildFile(pdir, &fi);
		}
	} while (_findnext(fh, &fi) >= 0);

out:
	_findclose(fh);
}

/*
 *	main
 */
int main(int argc, char* argv[])
{
	struct _stat	sb;
	long		size;
	DIROBJ		rootdir;

	/*
	 *	Check options
	 */

	argc--, argv++;
	if (argc <= 0)
		usage();

	while (argc && **argv == '-') switch (argv[0][1]) {
	default:
		usage();
		// NOTREACHED
	case 'o':
		argc--, argv++;
		if (argc <= 0)
			usage();
		outGBAfile = *argv;
		argc--, argv++;
		break;
	case 's':
		argc--, argv++;
		if (argc <= 0)
			usage();
		sysGBAfile = *argv;
		argc--, argv++;
		break;
	}

	if (argc <= 0)
		usage();

	strcpy(fsrootPath, *argv);

	if (_stat(fsrootPath, &sb) < 0)
		perror(fsrootPath), exit(1);
	if ((sb.st_mode & _S_IFMT) != _S_IFDIR)
		error("%s: is not a directory.", fsrootPath);

	if (_stat(sysGBAfile, &sb) < 0)
		perror(sysGBAfile), exit(1);
	if ((sb.st_mode & _S_IFMT) != _S_IFREG)
		error("%s: is not a file.", sysGBAfile);
	size = sb.st_size;

	if (!(sysfp = fopen(sysGBAfile, "rb")))
		perror(sysGBAfile), exit(1);

	if (!(outfp = fopen(outGBAfile, "w+b")))
		perror(outGBAfile), exit(1);

	/*
	 *	build file system
	 */

	fsCurOffset = 0;

	copyfile(sysGBAfile, sysfp, size);

	fsheadOffset = fsCurOffset;

	memset(&fshead, 0, sizeof(fshead));
	WriteFSData(&fshead, sizeof fshead);	// Reserved for fs header

	/*
	 *	build directories and files
	 */

	WalkDir(&rootdir, NULL, 1);

	BuildDirObj(&fshead.rootDir, &rootdir);

	/*
	 *	Build file system header
	 */

	strcpy(fshead.signature, ROMFSSIG);
	fshead.version = 0x01;
	fshead.size = sizeof(fshead);

	if (fseek(outfp, fsheadOffset, SEEK_SET))
		perror(outGBAfile), exit(1);
	WriteFSData(&fshead, sizeof fshead);

	/*
	 *	Finished
	 */

	FreeDirObj(&rootdir);

	fclose(outfp);
	fclose(sysfp);

	return 0;
}

